/****************************************************************************/
/*                  Copyright 1997, Trustees of Boston University.          */
/*                               All Rights Reserved.                       */
/*                                                                          */
/* Permission to use, copy, or modify this software and its documentation   */
/* for educational and research purposes only and without fee is hereby     */
/* granted, provided that this copyright notice appear on all copies and    */
/* supporting documentation.  For any other uses of this software, in       */
/* original or modified form, including but not limited to distribution in  */
/* whole or in part, specific prior permission must be obtained from Boston */
/* University.  These programs shall not be used, rewritten, or adapted as  */
/* the basis of a commercial software or hardware product without first     */
/* obtaining appropriate licenses from Boston University.  Boston University*/
/* and the author(s) make no representations about the suitability of this  */
/* software for any purpose.  It is provided "as is" without express or     */
/* implied warranty.                                                        */
/*                                                                          */
/****************************************************************************/
/*                                                                          */
/* Utility Name: Zipfs sequence generator                                   */
/* Authors:      Paul Barford                                               */
/* First Rev:    v1.0  12/13/96                                             */
/*                                                                          */
/*                                                                          */
/* Description:                                                             */
/*                                                                          */
/* This program enables the generate a sequence of numbers which follow     */
/* Zipfs law.  Zipfs law says that popularity is inversely proportional to  */
/* rank.  The sequence of numbers which are generated by this program are   */
/* then assigned to specific documents and thus are used to determine the   */
/* number of times that documents will be requested by SURGE.               */
/*                                                                          */
/* Inputs:                                                                  */
/*                                                                          */
/* The parameters for the program are listed in the DEFINE area of the      */
/* program header.  The program takes a two inputs from the command line.   */
/* The first is the maximum frequency number (ie. the number of times that  */
/* you want the most popular document to be requested).  This number must   */
/* be an integer.  The second is the total number of documents.  This       */
/* number must be an integer and must match the number specified when       */
/* sizes.c is run.                                                          */
/*                                                                          */
/* Run this program by typing the following:                                */
/*                                                                          */
/*                    zipf {max frequency} {num docs}                       */
/*                                                                          */
/* Outputs:                                                                 */
/*                                                                          */
/* The program outputs a single file called zout.txt.  It consists of two   */
/* columns, the first is the decending rank and the second are the decending*/
/* frequencies associated with each rank.                                   */
/*                                                                          */
/* NOTE:  The number of docs specified must be the same as the number       */
/* specified for sizes.c                                                    */
/*                                                                          */
/****************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#define MAXOUTPUT 100000     /* max number of ranked pairs generated        */

/* Define routines used in this program */

void initialize(char *, char *);
void results();

/* Define global variables used in this program */

long num_outfiles;          /* number of pairs written to arrays            */
long max_freq;              /* max frequency number from command line       */
long freq[MAXOUTPUT+1];     /* frequencies based on Zipfs relation          */

/****************************************************************************/
/* Main begins by simply initializing the output arrays and then applying   */
/* Zipfs law to the max number such that the inverse relation is held until */
/* the frequency is reduced to 1                                            */
/****************************************************************************/

void main(int argc, char *argv[])
{

  int i;
  long tot;
  float scale;

/****************************************************************************/
/* Initialize arrays.                                                       */
/****************************************************************************/

  if (argc < 2)
    {
      printf("Usage: {max freq value}, {num points} \n");
      exit(1);
    }
  
  initialize(argv[1],argv[2]);

/****************************************************************************/
/* We now calculate the frequencies                                         */
/****************************************************************************/

  printf("Generating reference values...\n");
  tot = max_freq;
  for(i=2;i<=num_outfiles;i++)
    {
      scale = 1.0/i;
      freq[i] = max_freq * scale;
      tot += freq[i];
    }
  printf("Total number of requests = %ld\n", tot);
 
/****************************************************************************/
/* Write out results.                                                       */
/****************************************************************************/

  results();

} /* End Main */

/****************************************************************************/
/* Initialize the data array                                                */
/****************************************************************************/

void initialize(char * maxfreq, char * numpoints)
{

  long i;

  max_freq =  atol(maxfreq);
  num_outfiles = atol(numpoints);

  if (num_outfiles > MAXOUTPUT)
    {
      printf("Must increase MAXOUTPUT \n");
      exit(1);
    }

/* Initialize the data array to 0 */

  for(i=1;i<=MAXOUTPUT;i++)
    {
      freq[i] = 0;
    }
  freq[1] = max_freq;

} /* End initialize */

/****************************************************************************/
/* The results routine opens zout.txt and prints out the rank, freq pairs   */
/****************************************************************************/

void results()
{
  long i;
  FILE *fp1;

  /* Open the output file zout.txt  */

  if((fp1 = fopen("zout.txt", "w")) == NULL)
    {
      perror("Unable to open zout.txt\n");
      exit(1);
    }

  /* Write results to output file     */

  for(i=1;i<=num_outfiles;i++)
    {
      fprintf(fp1,"%ld %ld \n", i, freq[i]);
    }

  /* Close the output file zout.txt */

  fclose(fp1);

} /* End results */




